from unittest import TestCase
from homeeasy.EncryptedMqtt import EncryptedMqtt


class TestEncryptedMqtt(TestCase):
    def test_get_key_1(self):
        self.assertEqual('a3d44c718acaead0', EncryptedMqtt._get_key('09ba20043c42').decode("utf-8"))

    def test_get_key_2(self):
        self.assertEqual('6e96059eb05bbf76', EncryptedMqtt._get_key('08bc270447e3').decode("utf-8"))

    def test_encrypt1(self):
        key = "a3d44c718acaead0".encode('utf-8')
        expected = '2776bf014dae8a1651994552b19e29969870167b550749d2362953a167fcc778'
        raw = bytes([170, 170, 18, 160, 10, 10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 26])
        mqtt = EncryptedMqtt()
        encrypt = mqtt._encrypt(raw, key)
        encrypted = encrypt.hex()
        self.assertEqual(expected, encrypted)

    def test_decrypt(self):
        key = "a3d44c718acaead0".encode('utf-8')
        encrypted = bytes.fromhex('a89488aaa5c79378e354a735e3f5cee2844b526b64d4a97ebf81f1d30e48e1b1')
        mqtt = EncryptedMqtt()
        decrypted = mqtt._decrypt(encrypted, key)
        expected = bytes.fromhex("aaaa12000a0a000c0700c4000000001c000000006d0000000000000000000000")
        self.assertEqual(expected, decrypted)



import asyncio
import logging
from typing import Any, Dict, Callable

import paho.mqtt.client as mqtt

from homeeasy.AsyncioHelper import AsyncioHelper
from homeeasy.CustomLogger import CustomLogger
from homeeasy.DeviceState import DeviceState
from homeeasy.EncryptedMqtt import EncryptedMqtt

logger = CustomLogger(logging.getLogger(__name__))


class HomeEasyLib:
    cache: Dict[str, DeviceState] = dict()
    status: Dict[str, Callable[[str, DeviceState], None]] = dict()
    _dump_status: Callable[[str, DeviceState], None] = None
    _dump_cmd: Callable[[str, DeviceState], None] = None
    mqtt_client: EncryptedMqtt

    def __init__(self) -> None:
        self.mqtt_client = EncryptedMqtt()
        self.mqtt_client.on_message = self.on_message

        loop = None
        try:
            loop = asyncio.events.get_running_loop()
        except RuntimeError:
            pass

        if loop is not None:
            AsyncioHelper(self.mqtt_client)

    def connect(self, host: str = "91.196.132.126", port: int = 1883):
        self.mqtt_client.connect(host, port)

        loop = None
        try:
            loop = asyncio.events.get_running_loop()
        except RuntimeError:
            pass

        if loop is None:
            self.mqtt_client.loop_start()

    def on_message(self, _client: mqtt.Client, _userdata: Any, mac: str, decrypted: bytes, message: mqtt.MQTTMessage):
        logger.debug("message received", topic=message.topic, payload=message.payload.hex(), decrypted=decrypted.hex())
        if len(decrypted) == 0:
            return
        state = DeviceState(decrypted)
        if 'dev/status/' in message.topic:
            self.cache[mac] = state
            if self._dump_status is not None:
                self._dump_status(mac, state)

            if mac in self.status:
                cb = self.status[mac]
                cb(mac, state)

        if 'dev/cmd/' in message.topic:
            if self._dump_cmd is not None:
                self._dump_cmd(mac, state)

    def dump_status(self, mac: str, cb: Callable[[str, DeviceState], None], topic_prefix: str = 'dev/status/010202/'):
        self._dump_status = cb
        topic = topic_prefix + mac
        self.mqtt_client.subscribe(topic)

    def dump_cmd(self, mac: str, cb: Callable[[str, DeviceState], None], topic_prefix: str = 'dev/cmd/010202/'):
        self._dump_cmd = cb
        topic = topic_prefix + mac
        self.mqtt_client.subscribe(topic)

    def disconnect(self):
        self.mqtt_client.disconnect()

    def request_status(self, mac: str, cb: Callable[[str, DeviceState], None],
                       cmd_topic_prefix: str = 'dev/cmd/010202/',
                       status_topic_prefix: str = 'dev/status/010202/'):
        self.status[mac] = cb

        data = bytes([170, 170, 18, 160, 10, 10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 26])
        self.mqtt_client.subscribe(status_topic_prefix + mac)
        self.mqtt_client.publish(cmd_topic_prefix + mac, data)

    def request_status_async(self, mac: str,
                             cmd_topic_prefix: str = 'dev/cmd/010202/',
                             status_topic_prefix: str = 'dev/status/010202/'):
        loop = asyncio.get_running_loop()
        fut = loop.create_future()

        self.request_status(mac, lambda _, state: fut.set_result(state), cmd_topic_prefix, status_topic_prefix)
        return fut

    def send(self, mac: str, status: DeviceState = None, topic_prefix: str = 'dev/cmd/010202/'):
        if status is None:
            if mac not in self.cache:
                return None
            status = self.cache[mac]
        self.mqtt_client.publish(topic_prefix + mac, status.cmd)

    def get(self, mac: str, key: str) -> Any:
        if mac not in self.cache:
            return None
        status = self.cache[mac]
        attr = getattr(status, key)
        return attr

    def set(self, mac: str, key: str, value: Any) -> Any:
        if mac not in self.cache:
            return None
        status = self.cache[mac]
        old = getattr(status, key)

        setattr(status, key, value)
        return old



import logging
from typing import Any, Callable, Dict
import paho.mqtt.client as mqtt
from Crypto.Cipher import AES
from Crypto.Hash import MD5
from Crypto.Util.Padding import pad

from homeeasy.CustomLogger import CustomLogger

logger = CustomLogger(logging.getLogger(__name__))


class EncryptedMqtt(mqtt.Client):
    _keys: Dict[str, bytes]

    def __init__(self, client_id="", clean_session=None, userdata=None,
                 protocol=mqtt.MQTTv311, transport="tcp"):
        super().__init__(client_id, clean_session, userdata, protocol, transport)
        self._keys = dict()

    def publish(self, topic, payload=None, qos=0, retain=False, properties=None):
        mac: str = self._mac_from_topic(topic)
        key = self._key_for_mac(mac)
        encrypted = self._encrypt(payload, key)
        return super().publish(topic, encrypted, qos, retain, properties)

    @property
    def on_message(self):
        return self._on_message_internal

    @on_message.setter
    def on_message(self, func: Callable[[mqtt.Client, Any, str, bytes, mqtt.MQTTMessage], None]):
        mqtt.Client.on_message.fset(self, func)

    def _on_message_internal(self, client, userdata, message):
        logger.debug("message received", topic=message.topic, payload=message.payload.hex())
        mac: str = self._mac_from_topic(message.topic)
        key = self._key_for_mac(mac)

        decrypted = self._decrypt(message.payload, key) if len(message.payload) > 0 else message.payload
        logger.debug("message decrypted", topic=message.topic, decrypted=decrypted.hex())
        super().on_message(client, userdata, mac, decrypted, message)

    def _key_for_mac(self, mac):
        if mac not in self._keys:
            key = self._get_key(mac)
            logger.debug("key generated", mac=mac, key=key.decode("utf-8"))
            self._keys[mac] = key
        else:
            key = self._keys[mac]
            logger.debug("key exists", mac=mac, key=key.decode("utf-8"))
        return key

    # noinspection PyMethodMayBeStatic
    def _mac_from_topic(self, topic):
        return topic.split('/')[-1]

    # noinspection PyMethodMayBeStatic
    def _decrypt(self, enc: bytes, key: bytes) -> bytes:
        iv = key[:AES.block_size]
        cipher = AES.new(key, AES.MODE_CBC, iv)
        dec = cipher.decrypt(enc)
        return dec  # unpad(dec, AES.block_size)

    @staticmethod
    def _encrypt(data: bytes, key: bytes) -> bytes:
        iv = key[:AES.block_size]
        cipher = AES.new(key, AES.MODE_CBC, iv)
        return cipher.encrypt(pad(data, AES.block_size))

    @staticmethod
    def _get_key(mac: str) -> bytes:
        magic_odd = '2Y10-6012-Y4'
        result = mac[::2]
        result += magic_odd

        h: MD5 = MD5.new()
        h.update(result.encode('utf-8'))
        result = h.hexdigest()[8: 24]

        return result.encode('utf-8')



from logging import Logger


class CustomLogger():

    def __init__(self, logger: Logger) -> None:
        super().__init__()
        self._logger = logger

    # noinspection PyMethodMayBeStatic
    def _extend_message(self, kwargs, msg):
        if len(kwargs) != 0:
            data_msg = str(kwargs)
            msg = msg + f"({data_msg})"
        return msg

    def debug(self, msg, *args, **kwargs) -> None:
        msg = self._extend_message(kwargs, msg)
        self._logger.debug(msg, *args, extra=kwargs)

    def info(self, msg, *args, **kwargs) -> None:
        msg = self._extend_message(kwargs, msg)
        self._logger.info(msg, *args, extra=kwargs)

    def warning(self, msg, *args, **kwargs) -> None:
        msg = self._extend_message(kwargs, msg)
        self._logger.warning(msg, *args, extra=kwargs)

    def warn(self, msg, *args, **kwargs) -> None:
        msg = self._extend_message(kwargs, msg)
        self._logger.warn(msg, *args, extra=kwargs)

    def error(self, msg, *args, **kwargs) -> None:
        msg = self._extend_message(kwargs, msg)
        self._logger.error(msg, *args, extra=kwargs)

    def exception(self, msg, *args, **kwargs) -> None:
        msg = self._extend_message(kwargs, msg)
        self._logger.exception(msg, *args, extra=kwargs)

    def critical(self, msg, *args, **kwargs) -> None:
        msg = self._extend_message(kwargs, msg)
        self._logger.critical(msg, *args, extra=kwargs)

    def log(self, level, msg, *args, **kwargs) -> None:
        msg = self._extend_message(kwargs, msg)
        self._logger.log(level, msg, *args,  extra=kwargs)



import asyncio
import logging

from paho.mqtt.client import Client, MQTT_ERR_SUCCESS

logger = logging.getLogger(__name__)


class AsyncioHelper:
    def __init__(self, client: Client):
        self.loop = asyncio.get_event_loop()
        self.client = client
        self.client.on_socket_open = self.on_socket_open
        self.client.on_socket_close = self.on_socket_close
        self.client.on_socket_register_write = self.on_socket_register_write
        self.client.on_socket_unregister_write = self.on_socket_unregister_write
        self.misc = None

    def on_socket_open(self, client, _userdata, sock):
        logger.debug("Socket opened")

        def cb():
            logger.debug("Socket is readable, calling loop_read")
            client.loop_read()

        self.loop.add_reader(sock, cb)
        self.misc = self.loop.create_task(self.misc_loop())

    def on_socket_close(self, _client, _userdata, sock):
        logger.debug("Socket closed")
        self.loop.remove_reader(sock)
        self.misc.cancel()

    def on_socket_register_write(self, client, _userdata, sock):
        logger.debug("Watching socket for writability.")

        def cb():
            logger.debug("Socket is writable, calling loop_write")
            client.loop_write()

        self.loop.add_writer(sock, cb)

    def on_socket_unregister_write(self, _client, _userdata, sock):
        logger.debug("Stop watching socket for writability.")
        self.loop.remove_writer(sock)

    async def misc_loop(self):
        logger.debug("misc_loop started")
        while self.client.loop_misc() == MQTT_ERR_SUCCESS:
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                break
        logger.debug("misc_loop finished")



import unittest
from datetime import time
from unittest import TestCase
from homeeasy.DeviceState import DeviceState, Mode, FanSpeed


class TestDeviceState(TestCase):
    data = bytes([170, 170, 18, 0, 10, 10, 0, 12, 7, 0, 196, 0, 0, 0, 0, 22, 5, 0, 0, 0, 108])

    # {"runMode": "100", "boot": 1, "windLevel": "000", "cpmode": 0, "mute": 0, "temtyp": 0, "wdNumber": 23,
    #  "windLR": "0000", "windTB": "0000", "lighting": 1, "healthy": 1, "timingMode": 0, "dryingmode": 0,
    #  "wdNumberMode": "01", "sleep": 0, "eco": 0, "bootEnabled": 0, "bootTime": "00:00", "shutEnabled": 0,
    #  "shutTime": "00:00", "wujiNum": 0, "indoorTemperature": "22.5", "windMode": 0}

    def test_run_mode(self):
        self.assertEqual(Mode.Heat, DeviceState(self.data).mode)

    def test_run_mode_set(self):
        state = DeviceState(self.data)
        state.mode = Mode.Fan
        self.assertEqual(Mode.Fan, state.mode)

    def test_boot(self):
        self.assertEqual(True, DeviceState(self.data).power)

    def test_boot_set(self):
        state = DeviceState(self.data)
        state.power = False
        self.assertEqual(False, state.power)

    def test_wind_level(self):
        self.assertEqual(FanSpeed.Auto, DeviceState(self.data).fanSpeed)

    def test_wind_level_set(self):
        state = DeviceState(self.data)
        state.fanSpeed = FanSpeed.l5
        self.assertEqual(FanSpeed.l5, state.fanSpeed)

    def test_cpmode(self):
        self.assertEqual(False, DeviceState(self.data).turbo)

    def test_cpmode_set(self):
        state = DeviceState(self.data)
        state.turbo = True
        self.assertEqual(True, state.turbo)

    def test_mute(self):
        self.assertEqual(False, DeviceState(self.data).quite)

    def test_mute_set(self):
        state = DeviceState(self.data)
        state.quite = True
        self.assertEqual(True, state.power)

    def test_temtyp(self):
        self.assertEqual(0, DeviceState(self.data).temperatureScale)

    def test_temtyp_set(self):
        state = DeviceState(self.data)
        state.temperatureScale = True
        self.assertEqual(True, state.temperatureScale)

    def test_wdNumber(self):
        self.assertEqual(23, DeviceState(self.data).desiredTemperature)

    def test_wdNumber_set(self):
        state = DeviceState(self.data)
        state.desiredTemperature = 17
        self.assertEqual(17, state.desiredTemperature)

    def test_windLR(self):
        self.assertEqual(0, DeviceState(self.data).flowHorizontalMode)

    def test_windLR_set(self):
        state = DeviceState(self.data)
        state.flowHorizontalMode = 3
        self.assertEqual(3, state.flowHorizontalMode)

    def test_windTB(self):
        self.assertEqual(0, DeviceState(self.data).flowVerticalMode)

    def test_windTB_set(self):
        state = DeviceState(self.data)
        state.flowVerticalMode = 3
        self.assertEqual(3, state.flowVerticalMode)

    def test_lighting(self):
        self.assertEqual(True, DeviceState(self.data).display)

    def test_lighting_set(self):
        state = DeviceState(self.data)
        state.display = False
        self.assertEqual(False, state.display)

    def test_healthy(self):
        self.assertEqual(True, DeviceState(self.data).healthy)

    def test_healthy_set(self):
        state = DeviceState(self.data)
        state.healthy = False
        self.assertEqual(False, state.healthy)

    def test_timingMode(self):
        self.assertEqual(0, DeviceState(self.data).timingMode)

    def test_timingMode_set(self):
        state = DeviceState(self.data)
        state.timingMode = True
        self.assertEqual(True, state.timingMode)

    def test_dryingmode(self):
        self.assertEqual(False, DeviceState(self.data).dryingMode)

    def test_dryingmode_set(self):
        state = DeviceState(self.data)
        state.dryingMode = True
        self.assertEqual(True, state.dryingMode)

    def test_wdNumberMode(self):
        self.assertEqual(1, DeviceState(self.data).wdNumberMode)

    def test_wdNumberMode_set(self):
        state = DeviceState(self.data)
        state.wdNumberMode = 3
        self.assertEqual(3, state.wdNumberMode)

    def test_sleep(self):
        self.assertEqual(False, DeviceState(self.data).sleep)

    def test_sleep_set(self):
        state = DeviceState(self.data)
        state.sleep = True
        self.assertEqual(True, state.sleep)

    def test_eco(self):
        self.assertEqual(False, DeviceState(self.data).eco)

    def test_eco_set(self):
        state = DeviceState(self.data)
        state.eco = True
        self.assertEqual(True, state.eco)

    def test_bootEnabled(self):
        self.assertEqual(False, DeviceState(self.data).bootEnabled)

    def test_bootEnabled_set(self):
        state = DeviceState(self.data)
        state.bootEnabled = True
        self.assertEqual(True, state.bootEnabled)

    def test_bootTime(self):
        self.assertEqual(time(0, 0), DeviceState(self.data).bootTime)

    def test_bootTime_set(self):
        state = DeviceState(self.data)
        state.bootTime = time(10, 12)
        self.assertEqual(time(10, 12), state.bootTime)

    def test_shutEnabled(self):
        self.assertEqual(False, DeviceState(self.data).shutEnabled)

    def test_shutEnabled_set(self):
        state = DeviceState(self.data)
        state.shutEnabled = True
        self.assertEqual(True, state.shutEnabled)

    def test_shutTime(self):
        self.assertEqual(time(0, 0), DeviceState(self.data).shutTime)

    def test_shutTime_set(self):
        state = DeviceState(self.data)
        state.shutTime = time(17, 18)
        self.assertEqual(time(17, 18), state.shutTime)

    def test_wujiNum(self):
        self.assertEqual(0, DeviceState(self.data).wujiNum)

    def test_wujiNum_set(self):
        state = DeviceState(self.data)
        state.wujiNum = 3
        self.assertEqual(3, state.wujiNum)

    def test_indoorTemperature(self):
        self.assertEqual(22.5, DeviceState(self.data).indoorTemperature)

    def test_windMode(self):
        self.assertEqual(0, DeviceState(self.data).fanMode)

    def test_windMode_set(self):
        state = DeviceState(self.data)
        state.fanMode = 7
        self.assertEqual(7, state.fanMode)

    def test_cmd(self):
        expected = b'\xaa\xaa\x12\x01\n\n\x00\x0c\x07\x00\xc4\x00\x00\x00\x00\x16\x05\x00\x00\x00m'
        self.assertEqual(expected, DeviceState(self.data).cmd)

    @unittest.skip("debug only")
    def test_str(self):
        s = str(DeviceState(self.data))
        self.assertEqual('', s)

    def test_mode_quite_set(self):
        state = DeviceState(self.data)
        state.mode = Mode.Fan
        state.quite = True
        self.assertEqual(Mode.Fan, state.mode)
        self.assertEqual(True, state.quite)



import setuptools

import codecs
import os.path


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setuptools.setup(
  name="homeeasy",
  version=get_version("homeeasy/__init__.py"),
  description="Control HVAC devices compatible with HomeEasy application.",
  long_description=read("README.md"),
  long_description_content_type="text/markdown",
  url="https://github.com/ki0ki0/homeeasylib",
  author="Ihor Kostiuk",
  author_email="igorkost@gmail.com",
  license="GPLv3",
  packages=setuptools.find_packages(),
  install_requires=["paho-mqtt", "pycryptodome", "structlog"],
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Operating System :: OS Independent",
    "Development Status :: 3 - Alpha",
    "Framework :: AsyncIO"
  ],
  python_requires='>=3.6',
)


from datetime import time
from enum import IntEnum
from typing import List, Dict

from homeeasy import valueHelper


class Mode(IntEnum):
    Auto = 0
    Cool = 1
    Dry = 2
    Fan = 3
    Heat = 4


class FanSpeed(IntEnum):
    Auto = 0
    l1 = 1
    l2 = 2
    l3 = 3
    l4 = 4
    l5 = 5
    l6 = 6


class FanMode(IntEnum):
    Auto = 0
    l1 = 1
    l2 = 2
    l3 = 3
    l4 = 4
    l5 = 5
    l6 = 6
    Quite = 7
    Turbo = 8


class HorizontalFlowMode(IntEnum):
    Stop = 0
    Swing = 1
    Left = 2
    Left_Center = 3
    Center = 4
    Right_Center = 5
    Right = 6
    Left_Right = 12
    Swing_Wide = 13


class VerticalFlowMode(IntEnum):
    Stop = 0
    Swing = 1
    Top = 2
    Top_Center = 3
    Center = 4
    Bottom_Center = 5
    Bottom = 6


class TemperatureScale(IntEnum):
    Celsius = 0
    Fahrenheit = 1


# noinspection PyPep8Naming
class DeviceState:
    bits: List[bool]

    _wd_number_dict: Dict[str, int] = {'00000': 61, '10000': 61, '00001': 62, '10001': 63, '00010': 64, '10010': 65,
                                      '00011': 66, '10011': 67, '00100': 68, '10100': 68, '00101': 69, '10101': 70,
                                      '00110': 71, '10110': 72, '00111': 73, '10111': 74, '01000': 75, '11000': 76,
                                      '01001': 77, '11001': 77, '01010': 78, '11010': 79, '01011': 80, '11011': 81,
                                      '01100': 82, '11100': 83, '01101': 84, '11101': 85, '01110': 86, '11110': 86,
                                      '01111': 87, '11111': 88}
    _wd_number_dict_invert = {value: key for key, value in _wd_number_dict.items()}

    def __init__(self, message: bytes) -> None:
        self.raw = message

    @staticmethod
    def _get_bit(x: int, b: int, size: int = 8) -> bool:
        return bool((x >> (size-1 - b)) & 1)

    def _bytes2bits(self, x: bytes) -> List[bool]:
        return [self._get_bit(x[int(pos / 8)], pos % 8) for pos in range(len(x) * 8)]

    @staticmethod
    def _bits2byte(bit0: bool, bit1: bool, bit2: bool, bit3: bool, bit4: bool, bit5: bool, bit6: bool, bit7: bool):
        byte_ = bit7 + (bit6 << 1) + (bit5 << 2) + (bit4 << 3) + (bit3 << 4) + (bit2 << 5) + (bit1 << 6) + (bit0 << 7)
        return byte_

    def _bits2bytes(self) -> bytes:
        bits = self.bits
        bytes_count = int(len(bits) / 8)
        list_ = [self._bits2byte(bits[pos * 8 + 0], bits[pos * 8 + 1], bits[pos * 8 + 2], bits[pos * 8 + 3],
                                 bits[pos * 8 + 4], bits[pos * 8 + 5], bits[pos * 8 + 6], bits[pos * 8 + 7]) for pos in
                 range(bytes_count)]
        return bytes(list_)

    def _get_state_bit(self, x: int, y: int) -> bool:
        x1 = x + 4
        return self.bits[x1 * 8 + y]

    def _set_state_bit(self, byte_pos: int, bit_pos: int, val: bool):
        byte_pos = byte_pos + 4  # offset on header size
        if type(val) is str:
            val_s = str(val).lower()
            val = False if val_s == 'false' or val_s == '0' else True
        self.bits[byte_pos * 8 + bit_pos] = bool(val)

    def _get_state_bits(self, x: int, y: int, count: int):
        val = 0
        for i in range(count):
            val = val << 1
            val += self._get_state_bit(x, y + i)
        return val

    def _set_state_bits(self, x: int, y: int, count: int, val: int):
        val = int(val)
        size = 8 * int((count / 8 + (1 if count % 8 > 0 else 0)))
        range_ = [self._get_bit(val, i, size) for i in range(size)]
        bits = range_[-count:]

        for i in range(count):
            self._set_state_bit(x, y + i, bool(bits[i]))

    # noinspection PyMethodMayBeStatic
    def _create_chunks(self, list_name, n):
        for i in range(0, len(list_name), n):
            yield list_name[i:i + n]

    def __repr__(self):
        keys = [i for i in dir(self) if not i.startswith('_')]
        skip = ['bits', 'cmd', 'raw']
        filtered = [key for key in keys if key not in skip]
        chunks = self._create_chunks(filtered, 6)

        chunks_with_values = [', '.join([f'{i}: {valueHelper.get_val(getattr(self, i))}' for i in j]) for j in chunks]
        out = ',\n'.join(chunks_with_values)
        return out

    def __str__(self):
        return self.__repr__()

    @property
    def raw(self) -> bytes:
        return self._bits2bytes()

    @raw.setter
    def raw(self, value: bytes):
        self.bits = self._bytes2bits(value[:21])

    @property
    def cmd(self) -> bytes:
        bits_bytes = bytearray(self._bits2bytes())
        bits_bytes[3] = 0x01
        code = 0x00
        for i in bits_bytes[:-1]:
            code += i
        bits_bytes[-1] = code & 0xFF
        return bytes(bits_bytes)

    @property
    def mode(self) -> Mode:
        return Mode(self._get_state_bits(3, 5, 3))

    @mode.setter
    def mode(self, value: Mode):
        if type(value) is str:
            s = str(value)
            try:
                value = Mode[s]
            except KeyError:
                value = Mode(int(s))
        self._set_state_bits(3, 5, 3, int(value))
            
    @property
    def power(self) -> bool:
        return self._get_state_bit(3, 4)

    @power.setter
    def power(self, value: bool):
        self._set_state_bit(3, 4, value)
            
    @property
    def fanSpeed(self) -> FanSpeed:
        return FanSpeed(self._get_state_bits(3, 1, 3))

    @fanSpeed.setter
    def fanSpeed(self, value: FanSpeed):
        if type(value) is str:
            s = str(value)
            try:
                value = FanSpeed[s]
            except KeyError:
                value = FanSpeed(int(s))
        self._set_state_bits(3, 1, 3, int(value))
            
    @property
    def turbo(self) -> bool:
        return self._get_state_bit(3, 0)

    @turbo.setter
    def turbo(self, value: bool):
        self._set_state_bit(3, 0, value)
    
    @property
    def quite(self) -> bool:
        return self._get_state_bit(4, 1)

    @quite.setter
    def quite(self, value: bool):
        self._set_state_bit(4, 1, value)
            
    @property
    def temperatureScale(self) -> TemperatureScale:
        return TemperatureScale(self._get_state_bit(4, 2))

    @temperatureScale.setter
    def temperatureScale(self, value: TemperatureScale):
        if type(value) is str:
            s = str(value)
            try:
                value = TemperatureScale[s]
            except KeyError:
                value = TemperatureScale(int(s))
        self._set_state_bit(4, 2, bool(value))
            
    @property
    def desiredTemperature(self) -> int:
        wen = self._get_state_bits(4, 3, 5)
        if not self.temperatureScale:
            wen = wen - 16 if wen >= 16 else wen
            wd_number = wen + 16 if wen > 0 else 16
        else:
            wen_str = format(wen, '#07b')[2:]
            wd_number = self._wd_number_dict[wen_str]
        return wd_number

    @desiredTemperature.setter
    def desiredTemperature(self, value: int):
        if not self.temperatureScale:
            val = value - 16 if value >= 16 else value
        else:
            val = self._wd_number_dict_invert[value];

        self._set_state_bits(4, 3, 5, val)
            
    @property
    def flowHorizontalMode(self) -> HorizontalFlowMode:
        return HorizontalFlowMode(self._get_state_bits(5, 0, 4))

    @flowHorizontalMode.setter
    def flowHorizontalMode(self, value: HorizontalFlowMode):
        if type(value) is str:
            s = str(value)
            try:
                value = HorizontalFlowMode[s]
            except KeyError:
                value = HorizontalFlowMode(int(s))
        self._set_state_bits(5, 0, 4, int(value))
            
    @property
    def flowVerticalMode(self) -> VerticalFlowMode:
        return VerticalFlowMode(self._get_state_bits(5, 4, 4))

    @flowVerticalMode.setter
    def flowVerticalMode(self, value: VerticalFlowMode):
        if type(value) is str:
            s = str(value)
            try:
                value = VerticalFlowMode[s]
            except KeyError:
                value = VerticalFlowMode(int(s))
        self._set_state_bits(5, 4, 4, int(value))
            
    @property
    def display(self) -> bool:
        return self._get_state_bit(6, 0)

    @display.setter
    def display(self, value: bool):
        self._set_state_bit(6, 0, value)
            
    @property
    def healthy(self) -> bool:
        return self._get_state_bit(6, 1)

    @healthy.setter
    def healthy(self, value: bool):
        self._set_state_bit(6, 1, value)
            
    @property
    def timingMode(self) -> bool:  # ?????
        return self._get_state_bit(6, 2)

    @timingMode.setter
    def timingMode(self, value: bool):
        self._set_state_bit(6, 2, value)
            
    @property
    def dryingMode(self) -> bool:  # Auxiliary heater (on, hot), Drying(off, cool or dry)
        return self._get_state_bit(6, 3)

    @dryingMode.setter
    def dryingMode(self, value: bool):
        self._set_state_bit(6, 3, value)
            
    @property
    def wdNumberMode(self) -> int:  # ??????
        return self._get_state_bits(6, 4, 2)

    @wdNumberMode.setter
    def wdNumberMode(self, value: int):
        self._set_state_bits(6, 4, 2, value)
            
    @property
    def sleep(self) -> bool:
        return self._get_state_bit(6, 6)

    @sleep.setter
    def sleep(self, value: bool):
        self._set_state_bit(6, 6, value)
            
    @property
    def eco(self) -> bool:
        return self._get_state_bit(6, 7)

    @eco.setter
    def eco(self, value: bool):
        self._set_state_bit(6, 7, value)
            
    @property
    def bootEnabled(self) -> bool:
        return self._get_state_bit(7, 4)

    @bootEnabled.setter
    def bootEnabled(self, value: bool):
        self._set_state_bit(7, 4, value)
            
    @property
    def bootTime(self) -> time:
        val_h = self._get_state_bits(7, 5, 3)
        val_l = self._get_state_bits(9, 0, 8)
        val = (val_h << 8) + val_l
        v_hour = int(val / 60)
        v_min = int(val % 60)
        return time(v_hour, v_min)

    @bootTime.setter
    def bootTime(self, value: time):
        value = self._str2time(value)
        val = value.hour * 60 + value.minute
        val_l = val & 0xff
        val_h = (val >> 8) & 0xff
        self._set_state_bits(7, 5, 3, val_h)
        self._set_state_bits(9, 0, 8, val_l)

    @staticmethod
    def _str2time(value):
        if type(value) is str:
            val_s = str(value)
            value = time.fromisoformat(val_s)
        return value

    @property
    def shutEnabled(self) -> bool:
        return self._get_state_bit(7, 0)

    @shutEnabled.setter
    def shutEnabled(self, value: bool):
        self._set_state_bit(7, 0, value)
            
    @property
    def shutTime(self) -> time:
        val_h = self._get_state_bits(7, 1, 3)
        val_l = self._get_state_bits(9, 0, 8)
        val = (val_h << 8) + val_l
        v_hour = int(val / 60)
        v_min = int(val % 60)
        return time(v_hour, v_min)

    @shutTime.setter
    def shutTime(self, value: time):
        value = self._str2time(value)
        val = value.hour * 60 + value.minute
        val_l = val & 0xff
        val_h = (val >> 8) & 0xff
        self._set_state_bits(7, 1, 3, val_h)
        self._set_state_bits(9, 0, 8, val_l)
            
    @property
    def wujiNum(self) -> int:  # ??????
        return self._get_state_bits(10, 0, 8)

    @wujiNum.setter
    def wujiNum(self, value: int):
        self._set_state_bits(10, 0, 8, value)
            
    @property
    def indoorTemperature(self) -> float:
        hi = self._get_state_bits(11, 0, 8)
        low = self._get_state_bits(12, 0, 8)
        return hi + 0.1 * low if not self.temperatureScale else hi * 1.3 + 32

    @property
    def fanMode(self) -> FanMode:
        wind_mode = int(self.fanSpeed)
        if self.turbo:
            wind_mode = 8
        else:
            if self.quite:
                wind_mode = 7
        return FanMode(wind_mode)

    @fanMode.setter
    def fanMode(self, value: FanMode):
        self.quite = False
        self.turbo = False

        if value == FanMode.Turbo:
            self.turbo = True
        elif value == FanMode.Quite:
            self.quite = True
        else:
            self.fanSpeed = FanSpeed(int(value))